/*! @file HexToC.c
 *
 *  A program to convert multiple Freescale S-record files into a C source file.
 *  Note: ignores values at or above \ref ICPFlashStart
 *
 *  Usage: HexToC \b outputFilename \b inputFilenameMask
 *
 *  Created on: 27/07/2008
 *      Author: PODonoghue
 */
#include <windows.h>
#include <stdio.h>
#include <io.h>
#include <time.h>
#include "Common.h"

U8 flashImage[0x10000];

#define ICPFlashStart (0xFC00) /*!< Start of Protected Flash memory \n
                                    Values at this address or above are ignored */

//! Struct used to contain information about the files converted
typedef struct {
   U8    hardwareVersion;  //!< hardware version of this image
   U8    softwareVersion;  //!< software version of this image
   U16   firstAddr;        //!< first non-0xFF address
   U16   lastAddr;         //!< last non-0xFF address
   char *sourceFilename;   //!< name of file loaded
} FlashImage;

//! Array used to contain information about the files converted
FlashImage flashImages[20];

//! Number of the last loaded file
int flashImageNumber;

/*! Convert a single character ('0'-'9', 'a'-'f' or 'A'-'F') into a hex value
 *
 * @param ptr  -  Ptr to the ptr to the character to convert. *ptr is advanced
 *
 * @return - a value in the range 0 to 15
 */
unsigned int hex1ToDecimal(char **ptr){
   U8 data = *(*ptr)++;
   if ((data >= '0') && (data <= '9'))
      return data - '0';
   if ((data >= 'a') && (data <= 'f'))
      return data - 'a' + 10;
   if ((data >= 'A') && (data <= 'F'))
      return data - 'A' + 10;
   return 0;
}

/*! Convert two characters ('0'-'9', 'a'-'f' or 'A'-'F') into a hex value
 *
 * @param ptr  -  Ptr to the ptr to the character to convert. *ptr is advanced
 *
 * @return - a value in the range 0 to 255
 */
unsigned int hex2ToDecimal( char **ptr){
   int data  = hex1ToDecimal(ptr);
   data      = data * 16 + hex1ToDecimal(ptr);
   return data;
}

/*! Convert four characters ('0'-'9', 'a'-'f' or 'A'-'F') into a hex value
 *
 * @param ptr  -  Ptr to the ptr to the character to convert. *ptr is advanced
 *
 * @return - a value in the range 0 to 65535
 */
unsigned int hex4ToDecimal( char **ptr){
   int data  = hex1ToDecimal(ptr);
   data      = data * 16 + hex1ToDecimal(ptr);
   data      = data * 16 + hex1ToDecimal(ptr);
   data      = data * 16 + hex1ToDecimal(ptr);
   return data;
}

/*!   Load a Freescale S-record file into the buffer
 *
 *    Minimal error checking (file open etc, no S-record validity checks). \n
 *    A global buffer \ref flashImage[] is used.
 *    The buffer is cleared to 0xFF before loading \n
 *    \ref flashImages[] is updated appropriately
 *
 *   @param filename - name (path) to the file to load
 *
 *   @return \n
 *           FALSE - load failed \n
 *           TRUE  - load was successful
 */
int loadSRecordFile( char *filename, U16 protectAdress ){
char *ptr;
unsigned int  addr;
unsigned int  size;
unsigned int  data;
FILE *fp;
char buffer[200];
U16  firstAddr = 0xFFFF;
U16  lastAddr  = 0x0000;

   fp = fopen(filename, "rt");

   if (fp == NULL) {
      fprintf(stderr, "Failed to open input file \"%s\"\n", filename);
      return FALSE;
   }
   fprintf(stderr, "Processing input file \"%s\"\n", filename);
   for (addr = 0; addr <= 0xFFFFUL; addr++)
      flashImage[addr] = 0xFF;
   while (fgets(buffer, sizeof(buffer)-1, fp) != NULL) {
      //fprintf(stderr, "Input:\"%s\"\n",buffer);
      ptr = buffer;
      if ((*ptr++ != 'S') || (*ptr++ != '1'))
         continue;
      size = hex2ToDecimal( &ptr );
      addr = hex4ToDecimal( &ptr );
      //fprintf(stderr, "(%2.2X)%4.4X:",size,addr);
      size -= 3;
      while (size-->0){
         data = hex2ToDecimal( &ptr );
         //fprintf(stderr, "%2.2X",data);
         if (addr >= protectAdress)
            continue;
         if (addr < firstAddr)
            firstAddr = addr;
         if (addr > lastAddr)
            lastAddr = addr;
         flashImage[addr++] = data;
      }
      //fprintf(stderr, "\n");
   }
   // Calculate were version information is in image
   U16 versionAddress = protectAdress-flashImage[protectAdress-2];
   flashImages[flashImageNumber].hardwareVersion = flashImage[versionAddress];
   flashImages[flashImageNumber].softwareVersion = flashImage[versionAddress+1];
   flashImages[flashImageNumber].lastAddr        = lastAddr;
   flashImages[flashImageNumber].firstAddr       = firstAddr;
   flashImages[flashImageNumber].sourceFilename  = strdup(filename);
   fclose(fp);
   return TRUE;
}

/*! Write buffer to C source file
 *
 *    @param fp - file pointer of open output file
 */
void writeCSourceFile( FILE *fp ){
unsigned int  addr;

   fprintf(fp, "//! Firmware image created from %s \n"
               "U8 flashImage%d[] = { ",
               flashImages[flashImageNumber].sourceFilename,
               flashImageNumber
               );

   if ((flashImages[flashImageNumber].firstAddr & 0x000F) != 0)
      fprintf(fp, "/* 0x%4.4X */ ", flashImages[flashImageNumber].firstAddr);

   for (addr =  flashImages[flashImageNumber].firstAddr;
        addr <= flashImages[flashImageNumber].lastAddr;
        addr++) {
      if ((addr & 0x000F) == 0x0)
         fprintf(fp, "/* 0x%4.4X */ ", addr);
      fprintf(fp,"0x%2.2X,",flashImage[addr]);
      if ((addr & 0x000F) == 0xF)
         fprintf(fp, "\n");
   }
   fprintf(fp, "};\n\n");
}

/*!   Converts multiple S-record files into a C source file
 *
 *    @param fp                  - handle for output file to write to
 *
 *    @param inputFilenameMask   - mask for input files.  This may contain a path.
 */
void convertFile(FILE *fp, U16 protectAdress, char *inputFilenameMask  ){
struct _finddata_t   fileinfo;
intptr_t             searchFileHandle=-1;
char inputDirectoryPath[MAX_PATH+1];
char *cp;

   // Get input folder
   strncpy(inputDirectoryPath,inputFilenameMask,MAX_PATH+1);
   inputDirectoryPath[MAX_PATH] = '\0';
   cp = strrchr(inputDirectoryPath,'\\');
   if (cp != NULL)
      *cp = '\0';
   fprintf(stderr, "Changing to directory \"%s\"\n", inputDirectoryPath);
   chdir(inputDirectoryPath);
   fprintf(stderr, "Matching input filenames \"%s\"\n", inputFilenameMask);
   searchFileHandle = _findfirst(inputFilenameMask, &fileinfo);
   if (searchFileHandle == -1)
      goto tidyUp;
   do {
      //printf("  %s\n", fileinfo.name);
      if (!loadSRecordFile(fileinfo.name, protectAdress))
         continue;
      writeCSourceFile( fp );
      if (++flashImageNumber>=sizeof(flashImages)/sizeof(flashImages[0])) { // Count this image
         fprintf(stderr, "Too many input files\n");
         goto tidyUp;
      }
   } while (_findnext(searchFileHandle, &fileinfo) == 0);


tidyUp:
   if (searchFileHandle != -1)
      _findclose(searchFileHandle);

}
/*!   Converts multiple S-record files into a C source file
 *
 *    @param outputFilename   - name of output file to create
 *
 *    @param inputFilenameMask   - mask for input files.  This may contain a path.
 */
void convertFiles(int  argc, char *argv[]) {

char           *outputFilename    = argv[1];
FILE           *fp;
unsigned int    imageNum = 0;
unsigned int    argCount;
time_t    time_now;

   time(&time_now);

   fp = fopen(outputFilename, "wt");

   if (fp == NULL) {
      fprintf(stderr, "Failed to open output file \"%s\"", outputFilename);
      goto tidyUp;
   }
   fprintf(stderr, "Producing output file \"%s\"\n", outputFilename);

   fprintf(fp,   "/*! \\file\n"
                 "    \\brief Firmware Flash Images for BDMs.\n\n"
                 "    Created %s"
                 "*/\n\n",
                 asctime(localtime(&time_now))
                 );

   for (argCount = 2; argCount < argc;){
      char *protectAddressS = argv[argCount++];
      char *inputFilenameMask = argv[argCount++];
      int  protectAddress;

      sscanf(protectAddressS, "%i", &protectAddress);
      convertFile(fp, protectAddress, inputFilenameMask );
   }

   fprintf(fp, "//! Structure for Firmware flash images\n");
   fprintf(fp,
           "typedef struct {\n"
           "   U8  hardwareVersion;    //!< Hardware version this image is intended for\n"
           "   U8  softwareVersion;    //!< Software version\n"
           "   U16 loadAddr;           //!< Address to load image at\n"
           "   U16 size;               //!< Size of image\n"
           "   U8  *data;              //!< Actual data\n"
           "   } FlashImage;\n\n"
           );

   fprintf(fp, "//! Firmware flash images for BDM, created %s//!\n",  asctime(localtime(&time_now)));
   fprintf(fp, "FlashImage flashImages[] = {\n"
               " // HW v  SW v  Load a   Size\n");
   for(imageNum = 0; imageNum<flashImageNumber; imageNum++)
      fprintf(fp,
           "   {0x%2.2X, 0x%2.2X, 0x%4X, 0x%4X, flashImage%d}, //!< Created from %s\n",
           flashImages[imageNum].hardwareVersion,
           flashImages[imageNum].softwareVersion,
           flashImages[imageNum].firstAddr,
           flashImages[imageNum].lastAddr-flashImages[imageNum].firstAddr+1,
           imageNum,
           flashImages[imageNum].sourceFilename
           );
   fprintf(fp, "};\n\n");

tidyUp:
   if (fp != NULL)
      fclose(fp);
}

int main(int  argc, char *argv[]){
   if (argc < 3) {
      fprintf(stderr, "Arguments(s) missing\n");
      fprintf(stderr, "Usage: hexToC outputFilename [protectedStartAddress inputFilenameMask]*\n");
      return 1;
   }
   convertFiles(argc, argv);
   return 0;
}
